@@ -53,6 +53,9 @@ class Graph:
                    repr(self.edges),
                    repr(self.title))
 
+    def _remove_cache_tag(self):
+        self.__dict__.pop('cache_tag', None)
+
     @staticmethod
     def has_unified_types(graphs):
         '''Check if all graphs have the same set of nodal/edge features.'''
@@ -89,6 +92,8 @@ class Graph:
         '''
 
         '''copy graphs if not editing in-place'''
+        for g in graphs:
+            g._remove_cache_tag()
         if inplace is not True:
             def shallowcopy(g):
                 h = cls(
@@ -120,7 +125,7 @@ class Graph:
         for component in ['nodes', 'edges']:
             group = [getattr(g, component) for g in graphs]
             for key in features[component]:
-                types = [g[key].get_type(concrete=True) for g in group]
+                types = [g[key].concrete_type for g in group]
                 t = common_min_type.of_types(types)
                 if t == np.object:
                     t = common_min_type.of_types(types, coerce=False)
