@@ -9,7 +9,7 @@ import numpy as np
 import sympy as sy
 from sympy.utilities.lambdify import lambdify
 from graphdot.codegen import Template
-from graphdot.codegen.typetool import cpptype
+from graphdot.codegen.cpptool import cpptype
 from graphdot.codegen.sympy_printer import cudacxxcode
 
 __all__ = ['BaseKernel',
@@ -17,13 +17,58 @@ __all__ = ['BaseKernel',
            'KroneckerDelta',
            'SquareExponential',
            'RationalQuadratic',
-           'TensorProduct']
+           'Normalize',
+           'Compose',
+           'TensorProduct',
+           'Additive',
+           'Convolution',
+           ]
 
 
 class BaseKernel:
 
+    def _assert_bounds(self, hyp, bounds):
+        if not isinstance(bounds, tuple) or len(bounds) != 2:
+            raise ValueError(
+                f'Bounds for hyperparameter {hyp} of kernel {self.name} must '
+                f'be a 2-tuple: {bounds} provided.'
+            )
+
     @staticmethod
-    def create(kernel, desc, expr, vars, *hyperparameter_specs):
+    def create(name, desc, expr, vars, *hyperparameter_specs):
+        '''Create a pairwise kernel class from a SymPy expression.
+
+        Parameters
+        ----------
+        name: str
+            The name of the kernel. Must be a valid Python identifier.
+        desc: str
+            A human-readable description of the kernel. Will be used to build
+            the docstring of the returned kernel class.
+        expr: str or SymPy expression
+            Expression of the kernel in SymPy format.
+        vars: 2-tuple of str or SymPy symbols
+            The input variables of the kernel as shown up in the expression.
+            A kernel must have exactly 2 input variables. All other symbols
+            that show up in its expression should be regarded as
+            hyperparameters.
+        hyperparameter_specs: list of hyperparameter specifications in one of
+        the formats below:
+
+            | symbol,
+            | (symbol,),
+            | (symbol, dtype),
+            | (symbol, dtype, description),
+            | (symbol, dtype, lower_bound, upper_bound),
+            | (symbol, dtype, lower_bound, upper_bound, description),
+
+            If a default set of lower and upper bounds are not defined here,
+            then it must be specified explicitly during kernel object
+            creation, using arguments as specified in the kernel class's
+            docstring.
+        '''
+
+        assert(isinstance(name, str) and name.isidentifier())
 
         '''parse expression'''
         if isinstance(expr, str):
@@ -77,14 +122,16 @@ class BaseKernel:
 
         class Kernel(BaseKernel, metaclass=CppType):
 
-            __name__ = kernel
-
             _expr = expr
             _vars = vars
             _hyperdefs = hyperdefs
             _dtype = np.dtype([(k, v['dtype']) for k, v in hyperdefs.items()],
                               align=True)
 
+            @property
+            def name(self):
+                return name
+
             def __init__(self, *args, **kwargs):
 
                 self._theta_values = values = OrderedDict()
@@ -100,7 +147,7 @@ class BaseKernel:
                         if symbol not in values:
                             raise KeyError(
                                 f'Hyperparameter {symbol} not provided '
-                                f'for {self.__name__}'
+                                f'for {self.name}'
                             )
 
                     try:
@@ -111,8 +158,10 @@ class BaseKernel:
                         except KeyError:
                             raise KeyError(
                                 f'Bounds for hyperparameter {symbol} of '
-                                f'kernel {self.__name__} not set, and no '
+                                f'kernel {self.name} not set, and '
+                                f'no defaults were given.'
                             )
+                    self._assert_bounds(symbol, bounds[symbol])
 
             # @cached_property
             @property
@@ -157,7 +206,7 @@ class BaseKernel:
 
             def __repr__(self):
                 return Template('${cls}(${theta, }, ${bounds, })').render(
-                    cls=self.__name__,
+                    cls=self.name,
                     theta=[f'{n}={v}' for n, v in self._theta_values.items()],
                     bounds=[f'{n}_bounds={v}'
                             for n, v in self._theta_bounds.items()]
@@ -190,7 +239,7 @@ class BaseKernel:
             @property
             def theta(self):
                 return namedtuple(
-                    self.__name__ + 'Hyperparameters',
+                    f'{self.name}Hyperparameters',
                     self._theta_values.keys()
                 )(**self._theta_values)
 
@@ -262,6 +311,19 @@ class BaseKernel:
             self,
         )
 
+    def __pow__(self, c):
+        r"""Implements the exponentiation semantics, i.e.
+        expression ``k1**c`` creates
+        :math:`k_{exp}(a, b) = k_1(a, b) ** k_2(a, b)`"""
+        if isinstance(c, (int, float)):
+            return KernelOperator.pow(self, Constant(c))
+        elif isinstance(c, BaseKernel) and c.name == 'Constant':
+            return KernelOperator.pow(self, c)
+        else:
+            raise ValueError(
+                f'Exponent must be a constant or constant kernel, got {c}.'
+            )
+
 
 class KernelOperator(BaseKernel):
     def __init__(self, k1, k2):
@@ -282,14 +344,20 @@ class KernelOperator(BaseKernel):
 
     @property
     def bounds(self):
-        return tuple(self.k1.bounds, self.k2.bounds)
+        return (self.k1.bounds, self.k2.bounds)
 
     @staticmethod
     def add(k1, k2):
         @cpptype(k1=k1.dtype, k2=k2.dtype)
         class Add(KernelOperator):
 
-            opstr = '+'
+            @property
+            def opstr(self):
+                return '+'
+
+            @property
+            def name(self):
+                return 'Add'
 
             def __call__(self, i, j, jac=False):
                 if jac is True:
@@ -314,9 +382,15 @@ class KernelOperator(BaseKernel):
     @staticmethod
     def mul(k1, k2):
         @cpptype(k1=k1.dtype, k2=k2.dtype)
-        class Mul(KernelOperator):
+        class Multiply(KernelOperator):
 
-            opstr = '*'
+            @property
+            def opstr(self):
+                return '*'
+
+            @property
+            def name(self):
+                return 'Multiply'
 
             def __call__(self, i, j, jac=False):
                 if jac is True:
@@ -345,7 +419,54 @@ class KernelOperator(BaseKernel):
                     f2 = self.k2.gen_expr(x, y, False, theta_prefix + 'k2.')
                     return f'({f1} * {f2})'
 
-        return Mul(k1, k2)
+        return Multiply(k1, k2)
+
+    @staticmethod
+    def pow(k1, k2):
+        @cpptype(k1=k1.dtype, k2=k2.dtype)
+        class Exponentiation(KernelOperator):
+
+            @property
+            def opstr(self):
+                return '**'
+
+            @property
+            def name(self):
+                return 'Exponentiation'
+
+            def __call__(self, i, j, jac=False):
+                if jac is True:
+                    # d(x^y) / dx = y * x^(y - 1)
+                    # d(x^y) / dy = x^y * log(x)
+                    f1, J1 = self.k1(i, j, True)
+                    f2, J2 = self.k2(i, j, True)
+                    return (
+                        f1**f2,
+                        np.array(
+                            [f2 * f1**(f2 - 1) * j1 for j1 in J1] +
+                            [f1**f2 * np.log(f1) * j2 for j2 in J2]
+                        )
+                    )
+                else:
+                    return self.k1(i, j, False)**self.k2(i, j, False)
+
+            def gen_expr(self, x, y, jac=False, theta_prefix=''):
+                if jac is True:
+                    f1, J1 = self.k1.gen_expr(x, y, True, theta_prefix + 'k1.')
+                    f2, J2 = self.k2.gen_expr(x, y, True, theta_prefix + 'k2.')
+                    return (
+                        f'__powf({f1}, {f2})',
+                        [f'({f2} * __powf({f1}, {f2} - 1) * {j})'
+                         for j in J1] +
+                        [f'(__powf({f1}, {f2}) * __logf({f1}) * {j})'
+                         for j in J2]
+                    )
+                else:
+                    f1 = self.k1.gen_expr(x, y, False, theta_prefix + 'k1.')
+                    f2 = self.k2.gen_expr(x, y, False, theta_prefix + 'k2.')
+                    return f'__powf({f1}, {f2})'
+
+        return Exponentiation(k1, k2)
 
 
 @cpptype([])
@@ -355,6 +476,10 @@ class _Multiply(BaseKernel):
     Only used internally for treating weighted edges
     """
 
+    @property
+    def name(self):
+        return '_Multiply'
+
     def __call__(self, x1, x2, jac=False):
         if jac is True:
             return x1 * x2, np.array([])
@@ -362,7 +487,7 @@ class _Multiply(BaseKernel):
             return x1 * x2
 
     def __repr__(self):
-        return '_Multiply()'
+        return f'{self.name}()'
 
     def gen_expr(self, x, y, jac=False, theta_prefix=''):
         f = f'({x} * {y})'
@@ -384,7 +509,7 @@ class _Multiply(BaseKernel):
         return tuple()
 
 
-def Constant(c, c_bounds=(0, np.inf)):
+def Constant(c, c_bounds=None):
     r"""Creates a no-op kernel that returns a constant value (often being 1),
     i.e. :math:`k_\mathrm{c}(\cdot, \cdot) \equiv constant`
 
@@ -398,12 +523,19 @@ def Constant(c, c_bounds=(0, np.inf)):
     BaseKernel
         A kernel instance of corresponding behavior
     """
+    if c_bounds is None:
+        c_bounds = (c, c)
 
     @cpptype(c=np.float32)
     class ConstantKernel(BaseKernel):
+        @property
+        def name(self):
+            return 'Constant'
+
         def __init__(self, c, c_bounds):
             self.c = float(c)
             self.c_bounds = c_bounds
+            self._assert_bounds('c', c_bounds)
 
         def __call__(self, i, j, jac=False):
             if jac is True:
@@ -412,7 +544,7 @@ def Constant(c, c_bounds=(0, np.inf)):
                 return self.c
 
         def __repr__(self):
-            return f'Constant({self.c})'
+            return f'{self.name}({self.c})'
 
         def gen_expr(self, x, y, jac=False, theta_prefix=''):
             f = f'{theta_prefix}c'
@@ -424,7 +556,7 @@ def Constant(c, c_bounds=(0, np.inf)):
         @property
         def theta(self):
             return namedtuple(
-                'ConstantHyperparameters',
+                f'{self.name}Hyperparameters',
                 ['c']
             )(self.c)
 
@@ -458,9 +590,14 @@ def KroneckerDelta(h, h_bounds=(1e-3, 1)):
     @cpptype(h=np.float32)
     class KroneckerDeltaKernel(BaseKernel):
 
+        @property
+        def name(self):
+            return 'KroneckerDelta'
+
         def __init__(self, h, h_bounds):
             self.h = float(h)
             self.h_bounds = h_bounds
+            self._assert_bounds('h', h_bounds)
 
         def __call__(self, i, j, jac=False):
             if jac is True:
@@ -472,7 +609,7 @@ def KroneckerDelta(h, h_bounds=(1e-3, 1)):
                 return 1.0 if i == j else self.h
 
         def __repr__(self):
-            return f'KroneckerDelta({self.h})'
+            return f'{self.name}({self.h})'
 
         def gen_expr(self, x, y, jac=False, theta_prefix=''):
             f = f'({x} == {y} ? 1.0f : {theta_prefix}h)'
@@ -484,7 +621,7 @@ def KroneckerDelta(h, h_bounds=(1e-3, 1)):
         @property
         def theta(self):
             return namedtuple(
-                'KroneckerDeltaHyperparameters',
+                f'{self.name}Hyperparameters',
                 ['h']
             )(self.h)
 
@@ -543,7 +680,7 @@ def Normalize(kernel):
     r"""Normalize a positive semidefinite kernel to produce the cosine of angle
     between the input variables.
     :math:`k_{N}(X, Y) = \frac{k_{base}(X, Y)}
-                              {\sqrt{k_{base}(X, X) k_{base}(Y, Y)}}`
+    {\sqrt{k_{base}(X, X) k_{base}(Y, Y)}}`
 
     Parameters
     ----------
@@ -553,6 +690,10 @@ def Normalize(kernel):
 
     @cpptype(kernel=kernel.dtype)
     class Normalized(BaseKernel):
+        @property
+        def name(self):
+            return 'Normalize'
+
         def __init__(self, kernel):
             self.kernel = kernel
 
@@ -578,7 +719,7 @@ def Normalize(kernel):
                 return Fxy * (Fxx * Fyy)**-0.5 if Fxx > 0 and Fyy > 0 else 0.0
 
         def __repr__(self):
-            return f'Normalize({repr(self.kernel)})'
+            return f'{self.name}({repr(self.kernel)})'
 
         def gen_expr(self, x, y, jac=False, theta_prefix=''):
             F, J = self.kernel.gen_expr(
@@ -606,7 +747,7 @@ def Normalize(kernel):
         @property
         def theta(self):
             return namedtuple(
-                'NormalizeHyperparameters',
+                f'{self.name}Hyperparameters',
                 ['kernel']
             )(self.kernel.theta)
 
@@ -625,8 +766,8 @@ def Compose(oper, **kw_kernels):
     r"""Creates a composite base kernel via the usage of a reduction operator
     to combine the outputs of multiple scalar kernels on individual features.
     :math:`k_\mathrm{composite}(X, Y; \mathrm{op}) =
-    k_{a_1}(X_{a_1}, Y_{a_1}) \mathrm{op} k_{a_2}(X_{a_2}, Y_{a_2}) \mathrm{op}
-    \ldots`
+    k_{a_1}(X_{a_1}, Y_{a_1})\,\mathrm{op}\,k_{a_2}(X_{a_2}, Y_{a_2})\,
+    \mathrm{op}\,\ldots`
 
     Parameters
     ----------
@@ -658,6 +799,10 @@ def Compose(oper, **kw_kernels):
 
     @cpptype([(key, ker.dtype) for key, ker in kw_kernels.items()])
     class Composite(BaseKernel):
+        @property
+        def name(self):
+            return 'Compose'
+
         def __init__(self, opstr, ufunc, jfunc, jgen, **kw_kernels):
             self.opstr = opstr
             self.ufunc = ufunc
@@ -666,7 +811,8 @@ def Compose(oper, **kw_kernels):
             self.kw_kernels = kw_kernels
 
         def __repr__(self):
-            return Template('Compose(${opstr}, ${kwexpr, })').render(
+            return Template('${cls}(${opstr}, ${kwexpr, })').render(
+                cls=self.name,
                 opstr=repr(self.opstr),
                 kwexpr=[f'{k}={repr(K)}' for k, K in self.kw_kernels.items()])
 
@@ -706,7 +852,7 @@ def Compose(oper, **kw_kernels):
         @property
         def theta(self):
             return namedtuple(
-                'CompositeHyperparameters',
+                f'{self.name}Hyperparameters',
                 self.kw_kernels.keys()
             )(*[k.theta for k in self.kw_kernels.values()])
 
@@ -759,6 +905,10 @@ def Convolution(kernel, normalize=True):
 
     @cpptype(kernel=kernel.dtype)
     class ConvolutionOf(BaseKernel):
+        @property
+        def name(self):
+            return 'Convolution'
+
         def __init__(self, kernel):
             self.kernel = kernel
 
@@ -772,7 +922,7 @@ def Convolution(kernel, normalize=True):
                 return np.sum([self.kernel(x, y) for x in X for y in Y])
 
         def __repr__(self):
-            return f'Convolution({repr(self.kernel)})'
+            return f'{self.name}({repr(self.kernel)})'
 
         def gen_expr(self, x, y, jac=False, theta_prefix=''):
             F, J = self.kernel.gen_expr(
@@ -799,7 +949,7 @@ def Convolution(kernel, normalize=True):
         @property
         def theta(self):
             return namedtuple(
-                'ConvolutionHyperparameters',
+                f'{self.name}Hyperparameters',
                 ['kernel']
             )(self.kernel.theta)
 
