@@ -1,6 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 import copy
+import itertools as it
 from collections import namedtuple
 import numpy as np
 from graphdot.graph import Graph
@@ -77,7 +78,7 @@ class MarginalizedGraphKernel:
         self.p = self._get_starting_probability(p)
         self.q = q
         self.q_bounds = q_bounds
-        self.dtype = dtype
+        self.element_dtype = dtype
 
         self.backend = backend_factory(backend)
 
@@ -107,10 +108,10 @@ class MarginalizedGraphKernel:
         lmin: 0 or 1
             Number of steps to skip in each random walk path before similarity
             is computed.
-            lmin + 1 corresponds to the starting value of l in the summation
+            (lmin + 1) corresponds to the starting value of l in the summation
             of Eq. 1 in Tang & de Jong, 2019 https://doi.org/10.1063/1.5078640
-            (or the first unnumbered equation as in Kashima, Tsuda, and
-            Inokuchi, 2003).
+            (or the first unnumbered equation in Section 3.3 of Kashima, Tsuda,
+            and Inokuchi, 2003).
 
         Returns
         -------
@@ -129,7 +130,8 @@ class MarginalizedGraphKernel:
         )
 
         ''' assert graph attributes are compatible with each other '''
-        pred_or_tuple = Graph.has_unified_types(X + Y if Y is not None else X)
+        all_graphs = list(it.chain(X, Y)) if Y is not None else X
+        pred_or_tuple = Graph.has_unified_types(all_graphs)
         if pred_or_tuple is not True:
             group, first, second = pred_or_tuple
             raise TypeError(
@@ -215,11 +217,11 @@ class MarginalizedGraphKernel:
 
         if traits.eval_gradient is True:
             return (
-                output[:, :, 0].astype(self.dtype),
-                output[:, :, 1:].astype(self.dtype)
+                output[:, :, 0].astype(self.element_dtype),
+                output[:, :, 1:].astype(self.element_dtype)
             )
         else:
-            return output.astype(self.dtype)
+            return output.astype(self.element_dtype)
 
     def diag(self, X, nodal=False, lmin=0, timing=False):
         """Compute the self-similarities for a list of graphs
@@ -237,10 +239,10 @@ class MarginalizedGraphKernel:
         lmin: 0 or 1
             Number of steps to skip in each random walk path before similarity
             is computed.
-            lmin + 1 corresponds to the starting value of l in the summation
+            (lmin + 1) corresponds to the starting value of l in the summation
             of Eq. 1 in Tang & de Jong, 2019 https://doi.org/10.1063/1.5078640
-            (or the first unnumbered equation as in Kashima, Tsuda, and
-            Inokuchi, 2003).
+            (or the first unnumbered equation in Section 3.3 of Kashima, Tsuda,
+            and Inokuchi, 2003).
 
         Returns
         -------
@@ -323,7 +325,7 @@ class MarginalizedGraphKernel:
             timer.report(unit='ms')
         timer.reset()
 
-        return output.astype(self.dtype)
+        return output.astype(self.element_dtype)
 
     """⭣⭣⭣⭣⭣ scikit-learn interoperability methods ⭣⭣⭣⭣⭣"""
 
